package main

import (
	"code.beefchicken.com/trimtools/stfw"
	"flag"
	"fmt"
	"io"
	"os"
	"path/filepath"
)

func main() {

	flag.Parse()

	if len(flag.Args()) != 2 {
		fmt.Printf("Usage:\n")
		fmt.Printf("    convertfw <input>.FW{1,2,3} <output prefix>\n")
		fmt.Printf("    convertfw <input>.bin{1,2,3} <output prefix>\n")
		os.Exit(1)
	}

	in := flag.Arg(0)
	out := flag.Arg(1)

	switch filepath.Ext(in) {
	case ".FW1", ".FW2", ".FW3":
		fw2bin(in, out)
	case ".bin1", ".bin2", ".bin3":
		bin2fw(in, out)
	default:
		die("Input file must have .bin{1,2,3} or .FW{1,2,3} extension. Aborting.")
	}
}

func bin2fw(binPath, fwPath string) {

	var area int
	var baseAddr uint32
	var startOffset uint32

	switch filepath.Ext(binPath) {
	case ".bin1":
		area = 1
		baseAddr = 0x24300000
		startOffset = 0x200
	case ".bin2":
		area = 2
		baseAddr = 0x24340000
	case ".bin3":
		area = 3
		baseAddr = 0x24380000
	default:
		die("Input file must have .bin1, .bin2 or .bin3 extension. Aborting.")
	}

	fwPathExt := fmt.Sprintf("%s.FW%d", fwPath, area)

	fmt.Printf("Loading binary data from '%s'...\n", filepath.Base(binPath))

	mem, err := os.ReadFile(binPath)
	if err != nil {
		die("Unable to read binary file '%s': %v. Aborting.\n", os.Args[1], err)
	}

	if len(mem)%32 != 0 {
		die("Binary must be a multiple of 32 bytes long. Aborting.")
	}

	fmt.Printf("Loaded successfully.\n")

	fmt.Printf("Verifying checksums...\n")

	scsl, scsh := stfw.GetChecksums(mem)
	ccsl, ccsh := stfw.CalcChecksums(mem, int(startOffset>>1))

	if scsl != ccsl {
		fmt.Printf("Low ROM Checksum mismatch: Expected 0x%02X, Found 0x%02X. Updating.", scsl, ccsl)
		stfw.UpdateChecksum(mem, stfw.MemoryModeCombinedLow, ccsl)
	}

	if scsh != ccsh {
		fmt.Printf("High ROM Checksum mismatch: Expected 0x%02X, Found 0x%02X. Updating.", scsh, ccsh)
		stfw.UpdateChecksum(mem, stfw.MemoryModeCombinedHigh, ccsh)
	}

	fmt.Printf("Writing image to '%s'...\n", filepath.Base(fwPathExt))

	f, err := os.Create(fwPathExt)
	if err != nil {
		die("Unable to create fw file '%s': %v. Aborting.\n", fwPathExt, err)
	}

	if area == 1 {
		fmt.Printf("Writing header for area 1 image...\n")

		// write an empty serial number table. It appears TFUPD.EXE doesn't check
		// this.
		if _, err := f.Write(make([]byte, 0x100)); err != nil {
			die("Error writing header: %v. Aborting.\n", err)
		}
	}

	fmt.Printf("Writing records...\n")

	for i := startOffset; i < uint32(len(mem)); i += 32 {
		err := stfw.WriteRecord(f, &stfw.Record{
			Addr: baseAddr + i,
			Data: mem[i : i+32],
		})
		if err != nil {
			die("Error writing record to file: %v. Aborting.\n", err)
		}
	}

	fmt.Printf("Write complete.\n")
	fmt.Printf("Records: %d\n", (len(mem)-int(startOffset))/32)
	fmt.Printf("Bytes: %d\n", len(mem)-int(startOffset))

	fmt.Printf("Address Range: 0x%08x - 0x%08x\n", baseAddr+startOffset, baseAddr+uint32(len(mem))-1)
}

func fw2bin(fwPath, binPath string) {

	var area int

	switch filepath.Ext(fwPath) {
	case ".FW1":
		area = 1
	case ".FW2":
		area = 2
	case ".FW3":
		area = 3
	default:
		die("Input file must have .FW1, .FW2 or .FW3 extension. Aborting.")
	}

	fmt.Printf("Loading image from '%s'...\n", filepath.Base(fwPath))

	f, err := os.Open(fwPath)
	if err != nil {
		die("Unable to open input file '%s': %v. Aborting.\n", fwPath, err)
	}

	if area == 1 {
		_, err = f.Seek(0x100, io.SeekStart)
		if err != nil {
			die("Unable seek past header: %v. Aborting.\n", err)
		}
	}

	mem := make([]byte, 0x20000)

	records := 0

	lowAddr := uint32(0xffffffff)
	highAddr := uint32(0)

	for {
		rec, err := stfw.ReadRecord(f)
		if err != nil {
			if err == io.EOF {
				break
			}
			die("Unexpected error reading input file '%s': %v. Aborting.\n", fwPath, err)
		}
		if !rec.Good {
			die("Bad checksum in record %d (Address 0x%08x). Aborting.\n", records, rec.Addr)
		}
		copy(mem[rec.Addr&0x1ffff:], rec.Data)
		records++
		if rec.Addr < lowAddr {
			lowAddr = rec.Addr
		}
		if rec.Addr > highAddr {
			highAddr = rec.Addr + 31
		}

	}
	fmt.Printf("Loaded successfully.\n")
	fmt.Printf("Records: %d\n", records)
	fmt.Printf("Bytes: %d\n", records*32)

	fmt.Printf("Address Range: 0x%08x-0x%08x\n", lowAddr, highAddr)

	binPathExt := fmt.Sprintf("%s.bin%d", binPath, area)
	fmt.Printf("Writing binary to '%s'...\n", filepath.Base(binPathExt))
	err = os.WriteFile(binPathExt, mem, 0644)
	if err != nil {
		die("Unable write binary file: %v. Aborting.\n", err)
	}

}

func die(format string, a ...interface{}) {
	fmt.Fprintf(os.Stderr, format, a...)
	os.Exit(1)
}
