/*

convertx - tools for mangling Trimble 4000 ROMs

Release 1.0

This is a collection of tools for working with fourth generation Trimble 4000
GPS receiver firmware images in the .X format as included with Trimble's
firmware update tool.

Because this code was written to solve a single instance of a single problem, it
is my tradition to write it as disposable code. This is code that is good enough
to get the job done, and not so terrible that I can't stand to look at it. But
no effort has been made to optimize it for memory or speed.

Another consequence of this style of development is that when I write tools such
as this, they are not stand-alone programs. I have made no effort to create a
command line interface to this code. I have structured it in a modular manner,
and I've used the Go paradigm of exported-functions-have-capital-letters to
identify the useful building blocks. I then use these building blocks in the
main() function to write a short program to accomplish the task at hand. If I
like it, it gets spun off into its own function to save it, which then frees up
my main() for the next project.

This program is in the public domain.

This program was written by Keelan Lightfoot
http://www.beefchicken.com/
*/

package main

import (
	"code.beefchicken.com/trimtools/dotx"
	"flag"
	"fmt"
	"io/ioutil"
	"os"
	"path/filepath"
	"strings"
)

const (
	memOpBlockSize = 244
)

func guessFileType(fn string) string {
	switch filepath.Ext(fn) {
	case ".X":
		return "x"
	case ".bin":
		return "bin"
	}
	return ""
}

func main() {

	var metaPath string

	var printInfo bool
	flag.BoolVar(&printInfo, "i", false, "Print image metadata")

	flag.Parse()

	if len(flag.Args()) != 2 {
		fmt.Printf("Usage:\n")
		fmt.Printf("    convertx <input>.X <output prefix>\n")
		fmt.Printf("    convertx <input>.bin <output prefix>\n")
		os.Exit(1)
	}

	inFile := flag.Arg(0)
	outFile := flag.Arg(1)

	fromBin := false
	switch filepath.Ext(inFile) {
	case ".X":
		metaPath = outFile + ".blocks"
		outFile = outFile + ".bin"
	case ".bin":
		fromBin = true
		outFile = outFile + ".X"
		metaPath = strings.TrimSuffix(inFile, filepath.Ext(inFile)) + ".blocks"
	default:
		die("Input file must have .bin or .X extension. Aborting.")
	}

	var dotX *dotx.DotXFile

	if fromBin {
		fmt.Printf("Loading block info from '%s'...\n", metaPath)

		meta, err := dotx.LoadMeta(metaPath)
		if err != nil {
			die("Unable to read block info from '%s': %v. Aborting.\n", metaPath, err)
		}

		fmt.Printf("Loading binary image from '%s'...\n", inFile)

		b, err := ioutil.ReadFile(inFile)
		if err != nil {
			die("Unable to read binary image from '%s': %v. Aborting.\n", inFile, err)
		}
		fmt.Printf("Combining block info and binary image...\n")

		dotX, err = dotx.NewDotXFromMetaAndMem(meta, b)
		if err != nil {
			die("Unable to combine block info and binary image: %v. Aborting.\n", err)
		}
	} else {
		fmt.Printf("Loading .X image from '%s'...\n", inFile)

		f, err := os.Open(inFile)
		if err != nil {
			die("Unable to open '%s': %v. Aborting.\n", inFile, err)
		}
		defer f.Close()
		dotX, err = dotx.ReadDotX(f)
		if err != nil {
			die("Unable to read .X image from '%s': %v. Aborting.\n", inFile, err)

		}
	}

	fmt.Printf("Loaded successfully.\n")

	mem := dotX.ToMemory()

	if printInfo {
		dotX.PrintInfo()
		mem.PrintInfo()
	}

	if false {
		// patch WNRO
		verMaj, verMin := mem.FirmwareVersion()
		ver := float64(verMaj) + float64(verMin)/100

		if ver < 7.19 {
			fmt.Printf("Warning: WRNO auto-patching has only been tested for firmware versions >= 7.19, use on other versions may produce unpredictable results.")
		}
		// Because the memory slice backs both blocks and records, the data can be modified
		// in any of the three records types simultaneously.
		// This will apply the WRNO patch to 732.X and 719.X
		// 1. find the matching instructions (addi.w  #$400,d2)
		matches := dotX.FindSequence([]uint8{0x06, 0x42, 0x04, 0x00})
		if len(matches) == 0 {
			fmt.Printf("Could not find ADDI.W  #$400,D2 instruction. Aborting WRNO patch attempt")
			os.Exit(1)
		}

		mem := matches[0].Block.Bytes
		match := matches[0].Matches[0]

		fmt.Printf("Found WNRO ADDI.W at %08X\n", matches[0].Block.StartAddr+match)

		// 2. change partial-rollover check cmpi to 0
		mem[match-4] = 0
		mem[match-3] = 0

		// 2. change $400 to $800
		mem[match+2] = 8

		// 3. recalculate the block checksums
		dotX.UpdateChecksums()
	}
	//
	// 	// Example 2: extracting and saving DSP firmware
	// 	if !strings.HasSuffix(os.Args[1], "732.X") {
	// 		log.Fatal("Example 2 only works with 732.X")
	// 	}
	//dotX.ExtractSigCode(0x7201C8),

	if fromBin {
		fmt.Printf("Updating checksums...")

		dotX.UpdateChecksums()

		fmt.Printf("Saving .X to '%s'...\n", outFile)
		out, err := os.Create(outFile)
		if err != nil {
			die("Unable write create .X file: %v. Aborting.\n", err)
		}
		defer out.Close()
		err = dotX.WriteDotX(out)
		if err != nil {
			die("Unable write .X file: %v. Aborting.\n", err)
		}
	} else {
		fmt.Printf("Saving block info to '%s'...\n", metaPath)
		err := dotX.SaveMeta(metaPath)
		if err != nil {
			die("Unable write block info file: %v. Aborting.\n", err)
		}
		fmt.Printf("Saving binary to '%s'...\n", outFile)
		err = dotX.SaveBinary(outFile)
		if err != nil {
			die("Unable write binary file: %v. Aborting.\n", err)
		}
	}

}

func die(format string, a ...interface{}) {
	fmt.Printf(format, a...)
	os.Exit(1)
}
