package main

import (
	"code.beefchicken.com/trimtools/dotx"
	"code.beefchicken.com/trimtools/monitor"
	"github.com/chzyer/readline"

	"encoding/binary"
	"flag"
	"fmt"
	"log"
	"math"
	"os"
	"strconv"
	"strings"
)

func main() {

	var baud int
	var debug, interactiveMonitor, showVersion, dummy bool
	var device, parity string
	var inFile, outFile string

	flag.IntVar(&baud, "b", 38400, "baud rate")
	flag.StringVar(&parity, "p", "none", "parity (none, even, lsodd)")

	flag.BoolVar(&debug, "x", false, "enable debug messages")
	flag.BoolVar(&interactiveMonitor, "m", false, "start interactive monitor")
	flag.BoolVar(&showVersion, "v", false, "show version and exit")
	flag.BoolVar(&dummy, "dummy", false, "dummy mode")
	flag.StringVar(&device, "d", "/dev/tty.usbserial-AB0JS0VU", "serial port device")
	flag.StringVar(&outFile, "s", "", "save firmware from receiver to file")
	flag.StringVar(&inFile, "f", "", "write firmware from file to receiver")
	flag.Parse()

	if showVersion {
		fmt.Println("loader(.exe) Trimble 4000 Firmware Loader")
		fmt.Println("Written by Keelan Lightfoot - http://beefchicken.com")
		return
	}

	var fw *dotx.DotXFile

	var mon *monitor.Monitor
	var rec *monitor.Receiver

	var err error

	if dummy {
		mon = &monitor.Monitor{
			//			Debug: debug,
			Dummy: dummy,
		}
		fmt.Println("Monitor dummy mode active.")

	} else {

		rec, err = monitor.Connect(device, baud, 8, 1, parity, debug)
		if err != nil {
			log.Fatal(err)
		}
		mon = monitor.NewMonitor(rec)
		defer mon.Close()
	}

	err = mon.StartMonitor()
	if err != nil {
		log.Fatal(err)
	}

	if interactiveMonitor {
		runMonitor(mon)
		return
	}

	if outFile != "" {
		fmt.Printf("Saving .X format firmware to '%s'...\n", outFile)
		fw, err = mon.CaptureReceiverFirmware()
		if err != nil {
			log.Fatal(err)
		}
		f, err := os.Create(outFile)
		if err != nil {
			log.Fatal(err)
		}
		err = fw.WriteDotX(f)
		if err != nil {
			log.Fatal(err)
		}
		f.Close()
	}

	if inFile != "" {
		fmt.Printf("Loading .X format firmware from '%s'...\n", inFile)
		f, err := os.Open(inFile)
		if err != nil {
			log.Fatal(err)
		}
		fw, err = dotx.ReadDotX(f)
		if err != nil {
			log.Fatal(err)
		}
		f.Close()
		err = mon.ProgramReceiver(fw)
		if err != nil {
			log.Fatal(err)
		}
	}

}

func runMonitor(mon *monitor.Monitor) {
	commands := map[string]func(*monitor.Monitor, []string) bool{
		"?":  handleHelpCommand,
		"q":  handleQuitCommand,
		"x":  handleQuitCommand,
		"r":  handleResetCommand,
		"u":  handleUCommand,
		"uw": handleUWCommand,
		"ul": handleULCommand,
		"ud": handleUDCommand,
		"d":  handleDCommand,
		"dw": handleDWCommand,
		"dl": handleDLCommand,
		"dd": handleDDCommand,
		"f":  handleFCommand,
		"v":  handleVersionCommand,
		"o":  handleOptionsCommand,
		"j":  handleJCommand,
	}

	rl, err := readline.New("enter command: ")
	if err != nil {
		panic(err)
	}
	defer rl.Close()

	for {
		fmt.Println("")
		s, err := rl.Readline()
		fmt.Println("")

		if err != nil {
			fmt.Println("bye")
			break
		}
		cmd := strings.Fields(s)
		if len(cmd) == 0 {
			continue
		}
		verb := strings.ToLower(cmd[0])

		if f := commands[verb]; f != nil {
			if f(mon, cmd) {
				break
			}
		} else {
			fmt.Printf("unrecognized command: %s\n", verb)
		}
	}

}

func handleQuitCommand(mon *monitor.Monitor, cmd []string) bool {
	return true
}

func handleHelpCommand(mon *monitor.Monitor, cmd []string) bool {
	fmt.Print(
		"Manual help mode:\n",
		"  Q or X       : Exit program.\n",
		"  R            : Reset target and exit program.\n",
		"  U addr count : hex/ascii dump of count bytes starting at addr\n",
		"                 (UD command works with doubles)\n",
		"                 (UW command works with words)\n",
		"                 (UL command works with longs)\n",
		"  D addr count data... : download data bytes to target address\n",
		"                 (DD command works with doubles)\n",
		"                 (DW command works with words)\n",
		"                 (DL command works with longs)\n",
		"  F addr count byte : fill target memory block\n",
		"  V            : Display firmware versions and code segment ranges.\n",
		"  O            : Display options data.\n",
		"  J addr       : Jump to address.\n",
		" \n",
		"  Addresses and data default to hex.  Counts default to decimal.\n",
		"  $xxxx or !dddd force hex or decimal.\n",
		"  CoMmAnDs ArE CaSe InSeNsITiVe.\n",
	)

	return false
}

func handleResetCommand(mon *monitor.Monitor, cmd []string) bool {
	mon.Stop()
	return false
}

func handleUCommand(mon *monitor.Monitor, cmd []string) bool {
	return ulCmd(mon, cmd, 1)
}

func handleUWCommand(mon *monitor.Monitor, cmd []string) bool {
	return ulCmd(mon, cmd, 2)
}

func handleULCommand(mon *monitor.Monitor, cmd []string) bool {
	return ulCmd(mon, cmd, 4)
}

func ulCmd(mon *monitor.Monitor, cmd []string, chunk int) bool {
	if len(cmd) < 3 {
		fmt.Printf("%s address count\n", cmd[0])
		return false
	}
	addr, ok := parseNum(cmd[1], 16)
	if !ok {
		fmt.Printf("invalid address: '%s'\n", cmd[1])
		return false
	}
	count, ok := parseNum(cmd[2], 10)
	if !ok {
		fmt.Printf("invalid count: '%s'\n", cmd[2])
		return false
	}

	b, err := mon.ReadMemory(addr, int(count)*chunk)
	if err != nil {
		log.Fatal(err)
	}

	var dataType string

	if chunk == 1 {
		dataType = "byte"
	} else if chunk == 2 {
		dataType = "word"
	} else if chunk == 4 {
		dataType = "long"
	}
	fmt.Printf("upload %s data from target memory\n", dataType)
	fmt.Printf("start address: %08x number of %ss: %d\n", addr, dataType, count)
	for i := 0; i < int(count)*chunk; i += 16 {
		e := i + 16
		if e > len(b) {
			e = len(b)
		}
		line := fmt.Sprintf("%08x ", addr+uint32(i))
		for q := 0; q < (e - i); q += chunk {
			if chunk == 1 {
				line = fmt.Sprintf("%s %02x", line, b[i+q])
			} else if chunk == 2 {
				line = fmt.Sprintf("%s %04x", line, binary.BigEndian.Uint16(b[i+q:]))
			} else if chunk == 4 {
				line = fmt.Sprintf("%s %08x", line, binary.BigEndian.Uint32(b[i+q:]))
			}
		}

		fmt.Printf("% -60s", line)
		for j := i; j < e; j++ {
			if b[j] >= 32 && b[j] <= 126 {
				fmt.Print(string(b[j]))
			} else {
				fmt.Print(".")
			}
		}
		fmt.Printf("\n")
	}
	return false
}

func handleUDCommand(mon *monitor.Monitor, cmd []string) bool {
	if len(cmd) < 3 {
		fmt.Println("u address count")
		return false
	}
	addr, ok := parseNum(cmd[1], 16)
	if !ok {
		fmt.Printf("invalid address: '%s'\n", cmd[1])
		return false
	}
	count, ok := parseNum(cmd[2], 10)
	if !ok {
		fmt.Printf("invalid count: '%s'\n", cmd[2])
		return false
	}
	b, err := mon.ReadMemory(addr, int(count)*8)
	if err != nil {
		fmt.Printf("read failed: %v\n", err)
	}
	fmt.Printf("upload doubles from target memory\n")
	fmt.Printf("start address: %08x number of doubles: %d\n", addr, count)

	for i := 0; i < int(count); i++ {
		fmt.Printf("%08x  %23.13e\n", int(addr)+i*8, math.Float64frombits(binary.BigEndian.Uint64(b[i*8:])))
	}
	return false
}

func handleDCommand(mon *monitor.Monitor, cmd []string) bool {
	return dlCmd(mon, cmd, 1)
}

func handleDWCommand(mon *monitor.Monitor, cmd []string) bool {
	return dlCmd(mon, cmd, 2)
}

func handleDLCommand(mon *monitor.Monitor, cmd []string) bool {
	return dlCmd(mon, cmd, 4)
}

func dlCmd(mon *monitor.Monitor, cmd []string, chunk int) bool {
	if len(cmd) < 3 {
		fmt.Printf("%s address v1 v2 v3 vn...\n", cmd[0])
		return false
	}
	addr, ok := parseNum(cmd[1], 16)
	if !ok {
		fmt.Printf("invalid address: '%s'\n", cmd[1])
		return false
	}

	var dataType string

	if chunk == 1 {
		dataType = "byte"
	} else if chunk == 2 {
		dataType = "word"
	} else if chunk == 4 {
		dataType = "long"
	}
	fmt.Printf("download %s data to target memory\n", dataType)
	fmt.Printf("start address: %08x number of %ss: %d\n", addr, dataType, len(cmd[2:]))

	v := make([]byte, len(cmd[2:])*chunk)
	i := 0
	for _, s := range cmd[2:] {
		l, ok := parseNum(s, 16)
		if !ok {
			fmt.Printf("invalid value: %s\n", s)
			return false
		}

		if chunk == 1 {
			v[i] = byte(l)
		} else if chunk == 2 {
			binary.BigEndian.PutUint16(v[i:], uint16(l))
		} else if chunk == 4 {
			binary.BigEndian.PutUint32(v[i:], l)
		}

		i += chunk
	}
	err := mon.WriteMemory(addr, v)
	if err != nil {
		fmt.Println("failed: %v\n", err)
	}
	return false
}

func handleDDCommand(mon *monitor.Monitor, cmd []string) bool {
	if len(cmd) < 3 {
		fmt.Println("dd address d1 d2 d3 dn...")
		return false
	}
	addr, ok := parseNum(cmd[1], 16)
	if !ok {
		fmt.Printf("invalid address: '%s'\n", cmd[1])
		return false
	}
	v := make([]byte, len(cmd[2:])*8)
	i := 0
	for _, s := range cmd[2:] {
		f, err := strconv.ParseFloat(s, 64)
		if err != nil {
			fmt.Printf("invalid float: %s\n", s)
			return false
		}

		binary.BigEndian.PutUint64(v[i:], math.Float64bits(f))
		i += 8
	}
	err := mon.WriteMemory(addr, v)
	if err != nil {
		fmt.Println("failed: %v\n", err)
	}
	return false
}

func handleFCommand(mon *monitor.Monitor, cmd []string) bool {
	if len(cmd) < 4 {
		fmt.Println("f address count byte")
		return false
	}
	addr, ok := parseNum(cmd[1], 16)
	if !ok {
		fmt.Printf("invalid address: '%s'\n", cmd[1])
		return false
	}
	count, ok := parseNum(cmd[2], 10)
	if !ok {
		fmt.Printf("invalid count: '%s'\n", cmd[2])
		return false
	}
	b, ok := parseNum(cmd[3], 16)
	if !ok {
		fmt.Printf("invalid byte: '%s'\n", cmd[3])
		return false
	}
	v := make([]byte, count)
	for i := 0; i < len(v); i++ {
		v[i] = byte(b)
	}
	err := mon.WriteMemory(addr, v)
	if err != nil {
		fmt.Println("failed: %v\n", err)
	}
	return false
}

func handleVersionCommand(mon *monitor.Monitor, cmd []string) bool {
	info, err := mon.GetReceiverInfo()
	if err != nil {
		fmt.Printf("GetReceiverInfo failed: %v\n", err)
		return false
	}
	info.PrintVersions()
	return false
}

func handleOptionsCommand(mon *monitor.Monitor, cmd []string) bool {
	info, err := mon.GetReceiverInfo()
	if err != nil {
		fmt.Printf("GetReceiverInfo failed: %v\n", err)
		return false
	}
	info.PrintOptions()
	return false
}

func handleJCommand(mon *monitor.Monitor, cmd []string) bool {
	if len(cmd) < 2 {
		fmt.Println("j address")
		return false
	}
	addr, ok := parseNum(cmd[1], 16)
	if !ok {
		fmt.Printf("invalid address: '%s'\n", cmd[1])
		return false
	}
	mon.CmdJump(addr)
	return false
}

func parseNum(s string, base int) (uint32, bool) {
	if strings.HasPrefix(s, "$") {
		s = strings.TrimPrefix(s, "$")
		base = 16
	} else if strings.HasPrefix(s, "!") {
		s = strings.TrimPrefix(s, "!")
		base = 10
	}

	v, err := strconv.ParseUint(s, base, 32)
	if err != nil {
		return 0, false
	}
	return uint32(v), true
}
