/*
dotx - tools for mangling Trimble 4000 ROMs

Release 1.0

This is a collection of tools for working with fourth generation Trimble 4000
GPS receiver firmware images in the .X format as included with Trimble's
firmware update tool.

Because this code was written to solve a single instance of a single problem, it
is my tradition to write it as disposable code. This is code that is good enough
to get the job done, and not so terrible that I can't stand to look at it. But
no effort has been made to optimize it for memory or speed.

Another consequence of this style of development is that when I write tools such
as this, they are not stand-alone programs. I have made no effort to create a
command line interface to this code. I have structured it in a modular manner,
and I've used the Go paradigm of exported-functions-have-capital-letters to
identify the useful building blocks. I then use these building blocks in the
main() function to write a short program to accomplish the task at hand. If I
like it, it gets spun off into its own function to save it, which then frees up
my main() for the next project.

This program is in the public domain.

This program was written by Keelan Lightfoot
http://www.beefchicken.com/
*/

package dotx

import (
	"bytes"
	"encoding/binary"
	"encoding/hex"
	"encoding/json"
	"fmt"
	"io"
	"io/ioutil"
	"log"
	"os"
	"strings"
	"time"
)

type DotXFile struct {
	ProductionFlag  uint8  `json:"productionFlag"`
	HeaderText      string `json:"headerText"`
	TransferAddress uint32 `json:"transferAddress"`
	DataBusWidth    uint16 `json:"dataBusWidth"`
	DataWidthBase   uint16 `json:"dataWidthBase"`
	// 	Memory          []byte
	Blocks []*Block `json:"blocks"`
	//	Records			[]*DataRecord
}

type Memory struct {
	Memory      []byte
	LowestAddr  uint32
	HighestAddr uint32
}

// The .X file format is a HP 64000 Absolute format file. Documentation of this
// format can be found starting on page 50 of this file:
//
// http://www.bitsavers.org/pdf/hp/64700/software_general/B1471-97008_Absolute_File_Translator_Mar92.pdf
//
// The above documentation shows that there is an 18 byte header in the .X
// format, but it does not describe the format of this header. From reverse-
// engineering the loader.exe, this is what I've determined about the format
// of that header:
//
// byte 0		(seems to be 0x82; actual value is ignored by loader.exe)
// byte 1		must be 0x04, is checked by loader.exe
// byte 2		(seems to be 0x38; actual value is ignored by loader.exe)
// byte 3-16	HP header (null terminated & padded). The length of this field
//				is a guess, loader.exe doesn't do any bounds checking.
// byte 17		firmware version
//		0x50 ('P') = production firmware
//		anything else = not production firmware
// If byte 17 indicates the file does not contain production firmware, the
// serial number authorization process is bypassed

func ReadDotX(f *os.File) (*DotXFile, error) {

	// Read Header (9 words)
	header := make([]byte, 18)
	_, err := f.Read(header)
	if err != nil {
		return nil, err
	}

	if header[1] != 0x04 {
		return nil, fmt.Errorf("bad header magic")
	}

	dotX := &DotXFile{}

	rawHeader := string(header[3:13])

	nullIndex := strings.Index(rawHeader, "\x00")
	if nullIndex >= 0 {
		dotX.HeaderText = rawHeader[:nullIndex]
	}

	pr := &processorRecord{}
	err = binary.Read(f, binary.BigEndian, pr)
	if err != nil {
		return nil, err
	}

	dotX.TransferAddress = uint32(pr.TransferAddressLS) + uint32(pr.TransferAddressMS)<<16
	dotX.ProductionFlag = header[17]
	dotX.DataBusWidth = pr.DataBusWidth
	dotX.DataWidthBase = pr.DataWidthBase
	// The HP docs say that if TransferAddressMS == 0, then the transfer address field is
	// to be ignored, so I don't actually do anything useful with the processor record.
	// Trimble native firmware seems to have the reset vector address (0x6BF0) in
	// TransferAddressLS, but in a .X generated by a firmware save, the field is all zeros.

	records := make([]*DataRecord, 0)
	for {
		dr := dataRecordHeader{}
		err = binary.Read(f, binary.BigEndian, &dr)
		if err != nil {
			if err == io.EOF {
				break
			}
			return nil, err
		}
		rec := &DataRecord{
			Addr:  uint32(dr.LoadAddressMS)<<16 + uint32(dr.LoadAddressLS),
			Bytes: make([]byte, dr.DataBytes),
		}

		_, err = f.Read(rec.Bytes)
		if err != nil {
			return nil, err
		}

		records = append(records, rec)
	}

	// figure out how big the target memory needs to be

	dotX.Blocks = []*Block{}

	var block *Block

	for _, r := range records {
		start := r.Addr
		end := r.Addr + uint32(len(r.Bytes)) - 1

		if block != nil {
			if start == block.EndAddr+1 { // continuing block
				block.EndAddr = end
				block.Bytes = append(block.Bytes, r.Bytes...)
			} else {
				dotX.Blocks = append(dotX.Blocks, block)
				block = nil
			}
		}
		if block == nil {
			block = &Block{
				StartAddr: start,
				EndAddr:   end,
				Bytes:     r.Bytes,
			}
		}

	}
	if block != nil {
		dotX.Blocks = append(dotX.Blocks, block)
	}

	return dotX, nil
}

func (dotX *DotXFile) ToMemory() *Memory {
	m := &Memory{}

	m.HighestAddr = uint32(0)
	m.LowestAddr = uint32(0xFFFFFFFF)

	// find memory bounds
	for _, block := range dotX.Blocks {
		if block.Checksummable() {
			if block.EndAddr > m.HighestAddr {
				m.HighestAddr = block.EndAddr
			}
			if block.StartAddr < m.LowestAddr {
				m.LowestAddr = block.StartAddr
			}
		}
	}

	m.Memory = make([]byte, m.HighestAddr+1)

	for _, block := range dotX.Blocks {
		if block.Checksummable() {
			copy(m.Memory[block.StartAddr:block.EndAddr+1], block.Bytes)
		}
	}

	return m
}

func LoadMeta(metaFN string) (*DotXFile, error) {
	b, err := ioutil.ReadFile(metaFN)
	if err != nil {
		return nil, err
	}
	dotX := &DotXFile{}
	if err = json.Unmarshal(b, dotX); err != nil {
		return nil, err
	}

	return dotX, nil
}

func NewDotXFromMetaAndMem(dotX *DotXFile, mem []byte) (*DotXFile, error) {

	for _, block := range dotX.Blocks {
		if block.Bytes == nil {
			block.Bytes = make([]byte, block.Len()+1)
			copy(block.Bytes, mem[block.StartAddr:block.EndAddr+1])
		}
	}
	dotX.UpdateChecksums()
	return dotX, nil

}

func NewDotX() *DotXFile {
	return &DotXFile{
		DataBusWidth:  16,
		DataWidthBase: 8,
		Blocks:        make([]*Block, 0),
	}
}

func (dotX *DotXFile) AddBlock(addr uint32, b []byte) {
	block := &Block{
		StartAddr: addr,
		EndAddr:   addr + uint32(len(b)),
		Bytes:     b,
	}
	dotX.Blocks = append(dotX.Blocks, block)
}

func (dotX *DotXFile) TotalBytes() int {
	t := 0
	for _, block := range dotX.Blocks {
		t += len(block.Bytes)
	}
	return t
}

func (dotX *DotXFile) WriteDotX(f *os.File) error {

	// Write Trimble Header
	trimbleHeader := make([]byte, 18)
	trimbleHeader[0] = 0x82 // ?
	trimbleHeader[1] = 0x04 // magic number
	trimbleHeader[2] = 0x38 // 0x00 for 4000, 0x38 for 4400?
	copy(trimbleHeader[3:17], dotX.HeaderText)
	trimbleHeader[17] = dotX.ProductionFlag // Production Version
	f.Write(trimbleHeader)

	// Write Processor Record
	pr := &processorRecord{
		Length:            7,
		DataBusWidth:      dotX.DataBusWidth,
		DataWidthBase:     dotX.DataWidthBase,
		TransferAddressLS: uint16(dotX.TransferAddress & 0xFFFF),
		TransferAddressMS: uint16(dotX.TransferAddress >> 16 & 0xFFFF),
	}
	err := binary.Write(f, binary.BigEndian, pr)
	if err != nil {
		return err
	}
	// cs1 := &DataRecord{
	// 		Addr:  0x00FFFA54,
	// 		Bytes: []byte{0x70, 0x07, 0x78, 0x71},
	// 	}
	// 	err = cs1.Write(f)
	// 	if err != nil {
	// 		return err
	// 	}
	// 	cs2 := &DataRecord{
	// 		Addr:  0x00FFFA46,
	// 		Bytes: []byte{0x03, 0xF5},
	// 	}
	// 	err = cs2.Write(f)
	// 	if err != nil {
	// 		return err
	// 	}

	for _, block := range dotX.Blocks {
		err = block.Write(f)
		if err != nil {
			return err
		}
	}

	return nil
}

func WriteBinary(memory []byte, outfile string) error {
	err := ioutil.WriteFile(outfile, []byte(memory), 0644)
	return err
}

func (dotX *DotXFile) SaveBinary(outfile string) error {
	memory := dotX.ToMemory()
	err := ioutil.WriteFile(outfile, []byte(memory.Memory), 0644)
	if err != nil {
		return err
	}

	return err

}

func (dotX *DotXFile) SaveMeta(outfile string) error {

	clone := *dotX
	clone.Blocks = make([]*Block, len(dotX.Blocks))

	for i, block := range dotX.Blocks {
		b := *block
		if block.Checksummable() {
			b.Bytes = nil
		}
		clone.Blocks[i] = &b
	}

	b, err := json.MarshalIndent(clone, "", "  ")
	if err != nil {
		return err
	}
	err = ioutil.WriteFile(outfile, b, 0644)

	return err

}

func (m *Memory) ExtractSigCode(baseAddr uint32) []byte {

	tmsRAM := make([]byte, 0x8000)

	inAddr := baseAddr + 0x1C

	// 0-1		record length - 1 (0 = stop)
	// 2-3		number of bytes in record
	// 4-5		destination offset
	// 6...n	data
	for {
		if binary.BigEndian.Uint16(m.Memory[inAddr:inAddr+3]) == 0 {
			break
		}
		i := binary.BigEndian.Uint16(m.Memory[inAddr+2:inAddr+5]) >> 1
		outAddr := uint32(binary.BigEndian.Uint16(m.Memory[inAddr+4:inAddr+7])) << 1
		inAddr += 8
		for ; i > 0; i-- {
			// TMS320C25 is little endian
			// MC68332 is big endian
			tmsRAM[outAddr] = m.Memory[inAddr+1]
			tmsRAM[outAddr+1] = m.Memory[inAddr]
			inAddr += 2
			outAddr += 2

		}

	}
	return tmsRAM
}

func (dotX *DotXFile) PrintInfo() {

	fmt.Println("----------------- Header Record -----------------")
	fmt.Printf("        Production firmware: %v\n", dotX.ProductionFlag == 'P')
	fmt.Printf("        HP64000 header name: %s\n", dotX.HeaderText)

	fmt.Println("------------ Processor Record ------------")
	fmt.Printf("             Data Bus Width: %d\n", dotX.DataBusWidth)
	fmt.Printf("            Data Width Base: %d\n", dotX.DataWidthBase)
	fmt.Printf("           Transfer Address: 0x%08X\n", dotX.TransferAddress)

	dotX.PrintBlocks()
}

func (m *Memory) PrintInfo() {

	fmt.Println("-------------- Data Records --------------")
	fmt.Printf("             Lowest Address: 0x%08X\n", m.LowestAddr)
	fmt.Printf("            Highest Address: 0x%08X\n", m.HighestAddr)

	fmt.Println("-------------- Firmware Metadata --------------")
	verMaj, verMin := m.FirmwareVersion()
	fmt.Printf("           Firmware Version: %d.%d\n", verMaj, verMin)
	fmt.Printf("              Test firmware: %v\n", m.Memory[0x229] != 0)
	fmt.Printf("              Firmware Date: %v\n", m.FirmwareDate().Format("02-Jan-2006"))

}

func (m *Memory) FirmwareVersion() (int, int) {
	ver := int(binary.BigEndian.Uint16(m.Memory[0x204 : 0x205+1]))
	major := ver / 100
	minor := ver % 100
	return major, minor
}

func (m *Memory) FirmwareDate() time.Time {

	date := int(binary.BigEndian.Uint32(m.Memory[0x222 : 0x222+5]))
	month := date / 10000
	day := date / 100 % 100
	year := date % 100

	if year >= 80 {
		year += 1900
	} else {
		year += 2000
	}

	t := time.Date(year, time.Month(month), day, 0, 0, 0, 0, time.UTC)

	return t
}

func (dotX *DotXFile) PrintBlocks() {

	for i, block := range dotX.Blocks {
		fmt.Printf("-------------------- Block %d --------------------\n", i)
		fmt.Printf("                      Start: %08X\n", block.StartAddr)
		fmt.Printf("                        End: %08X\n", block.EndAddr)

		if block.Checksummable() {
			cs, _ := block.Checksum()
			fmt.Printf("                   Checksum: %08X\n", cs)
			ccs, _ := block.CalculateChecksum()
			fmt.Printf("                 Calculated: %08X\n", ccs)
		} else {
			fmt.Printf("                      Bytes: % 02X\n", block.Bytes)
		}
	}

}

type SequenceMatch struct {
	Block   *Block
	Matches []uint32
}

func (dotX *DotXFile) FindSequence(seq []byte) []SequenceMatch {

	matches := make([]SequenceMatch, 0)
	for _, block := range dotX.Blocks {
		blockMatches := make([]uint32, 0)
		for i := 0; i <= len(block.Bytes)-len(seq); i++ {
			if bytes.Compare(block.Bytes[i:i+len(seq)], seq) == 0 {
				blockMatches = append(blockMatches, uint32(i)+block.StartAddr)
			}
		}
		if len(blockMatches) > 0 {
			matches = append(matches, SequenceMatch{
				Block:   block,
				Matches: blockMatches,
			})
		}
	}
	return matches

}

func (dotX *DotXFile) UpdateChecksums() {
	for _, block := range dotX.Blocks {
		block.UpdateChecksum()
	}
}

type processorRecord struct {
	Length            uint16
	DataBusWidth      uint16
	DataWidthBase     uint16
	TransferAddressLS uint16
	TransferAddressMS uint16
}

// func (p *processorRecord) TransferAddress() int {
// 	return int(p.TransferAddressLS) + int(p.TransferAddressMS)<<16
// }

type dataRecordHeader struct {
	Length        uint16
	DataBytes     uint16
	LoadAddressLS uint16
	LoadAddressMS uint16
}

type DataRecord struct {
	Addr  uint32
	Bytes []byte
	//Header  dataRecordHeader
}

func (d *DataRecord) LastAddr() uint32 {
	return d.Addr + uint32(len(d.Bytes)) - 1
}

func (d *DataRecord) Write(f *os.File) error {
	header := dataRecordHeader{
		Length:        uint16(len(d.Bytes) + 5),
		DataBytes:     uint16(len(d.Bytes)),
		LoadAddressMS: uint16(d.Addr >> 16 & 0xFFFF),
		LoadAddressLS: uint16(d.Addr & 0xFFFF),
	}
	err := binary.Write(f, binary.BigEndian, header)
	if err != nil {
		return err
	}
	err = binary.Write(f, binary.BigEndian, d.Bytes)
	if err != nil {
		return err
	}
	return nil
}

type hexByte []byte

func (w *hexByte) MarshalJSON() ([]byte, error) {
	return []byte(`"` + hex.EncodeToString(*w) + `"`), nil
}

func (w *hexByte) UnmarshalJSON(b []byte) error {
	b, err := hex.DecodeString(string(b[1 : len(b)-1]))
	if err != nil {
		return err
	}
	*w = b
	return nil
}

type Block struct {
	StartAddr uint32  `json:"startAddr"`
	EndAddr   uint32  `json:"endAddr"`
	Bytes     hexByte `json:"data,omitempty"`
}

func (b *Block) Write(f *os.File) error {
	if b.Checksummable() {
		err := b.UpdateChecksum()
		if err != nil {
			return err
		}
	}
	for _, rec := range b.ToRecords(244) {
		err := rec.Write(f)
		if err != nil {
			return err
		}
	}
	return nil
}

func (b *Block) Len() int {
	return int(b.EndAddr - b.StartAddr)
}

func (b *Block) ToRecords(blockSize int) []*DataRecord {
	recs := make([]*DataRecord, 0)
	length := len(b.Bytes)
	for i := 0; i < length; i += blockSize {
		l := blockSize
		if i+l > length {
			l = length - i
		}
		recs = append(recs, &DataRecord{
			Addr:  uint32(i) + b.StartAddr,
			Bytes: b.Bytes[i : i+l],
		})
	}
	return recs
}

func (b *Block) Checksummable() bool {
	return b.EndAddr-b.StartAddr+1 >= 12

}

func (b *Block) Checksum() (uint32, error) {

	if !b.Checksummable() {
		return 0, fmt.Errorf("block too small for checksum")
	}

	p := len(b.Bytes) - 12

	return binary.BigEndian.Uint32(b.Bytes[p : p+4]), nil

}

func (b *Block) CalculateChecksum() (uint32, error) {

	if !b.Checksummable() {
		return 0, fmt.Errorf("block too small for checksum")
	}

	p := len(b.Bytes) - 12
	reader := bytes.NewReader(b.Bytes)

	var checksum uint32
	for i := 0; i < p/4; i++ {
		var v uint32
		err := binary.Read(reader, binary.BigEndian, &v)
		if err != nil {
			log.Fatal(err)
		}
		checksum += v
	}

	return checksum, nil
}

//func (b *Block) UpdateChecksum() error {

func (b *Block) UpdateChecksum() error {

	if !b.Checksummable() {
		return fmt.Errorf("block too small for checksum")
	}

	checksum, err := b.CalculateChecksum()
	if err != nil {
		return err
	}

	p := len(b.Bytes) - 12
	binary.BigEndian.PutUint32(b.Bytes[p:p+4], checksum)

	return nil
}
